/*
 *  fakewww [port]
 *   concurrent server                          thd@ornl.gov
 * reap children, and thus watch out for EINTR
 *  take http GET / and return a web page, then return requested file by
 *   next GET
 * can use this to "provide" a java client for machine without having
 * to run a web server
 */

#include        <stdio.h>
#include        <sys/types.h>
#include        <sys/socket.h>
#include        <netinet/in.h>
#include        <arpa/inet.h>
#include        <netdb.h>
#include        <signal.h>
#include        <string.h>


#define PORT 7123
#define BUFFSIZE        1024
char buff[BUFFSIZE];
/* html message */
char *MsgOK  = "HTTP/1.0 200 OK\n\n";
char *MsgNope = "HTTP/1.0 404 Not found\n\n"
    "<HEAD><TITLE>File Not Found</TITLE></HEAD>\n"
        "<BODY><H1>The requested file could not be found</H1></BODY>\n";
char *Mypagefile = "/tcpbw100.html";   /* we throw the slash away */
char *okfile[] = {"/tcpbw100.html", "/Tcpbw100.class",0};

#include <sys/errno.h>
extern int errno;


main(argc, argv)
int	argc;
char	*argv[];
{
	int port = PORT, one = 1;
	void reap();
	int n,sockfd, newsockfd, clilen, childpid, servlen;
	struct sockaddr_in	cli_addr, serv_addr;
	FILE *fp;


	if (argc > 1) port = atoi(argv[1]);
	if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		err_sys("server: can't open stream socket");

	/*
	 * Bind our local address so that the client can send to us.
	 */

	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port        = htons(port);

	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
	if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		err_sys("server: can't bind local address");

	clilen = sizeof(cli_addr);
	getsockname(sockfd, (struct sockaddr *)&cli_addr, &clilen);
	printf("server ready on port %d\n",ntohs(cli_addr.sin_port));
	listen(sockfd, 5);
	signal(SIGCLD,reap);		/* get rid of zombies */

		/*
		 * Wait for a connection from a client process.
		 * This is an example of a concurrent server.
		 */

	for(;;){
		int i;

		clilen = sizeof(cli_addr);
		newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
		if (newsockfd < 0){
			if (errno == EINTR) continue; /*sig child */
			err_sys("server: accept error");
		}

		if (fork() == 0){ /* child */
			close(sockfd);
			dowww(newsockfd);
			exit(0);
		}
		close(newsockfd);
	}

}

#include        <sys/wait.h>
void reap()
{
        /*
	 * avoid zombies, since we run forever
         * Use the wait3() system call with the WNOHANG option.
         */

        int             pid;
        union wait      status;

        while ( (pid = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0)
                ;
}

err_sys(s)
char *s;
{
	perror(s);
	exit(1);
}
/*
 * Read a line from a descriptor.  Read the line one byte at a time,
 * looking for the newline.  We store the newline in the buffer,
 * then follow it with a null (the same as fgets(3)).
 * We return the number of characters up to, but not including,
 * the null (the same as strlen(3)).
 */

int
readline(fd, ptr, maxlen)
register int	fd;
register char	*ptr;
register int	maxlen;
{
	int	n, rc;
	char	c;

	for (n = 1; n < maxlen; n++) {
		if ( (rc = read(fd, &c, 1)) == 1) {
			*ptr++ = c;
			if (c == '\n')
				break;
		} else if (rc == 0) {
			if (n == 1)
				return(0);	/* EOF, no data read */
			else
				break;		/* EOF, some data was read */
		} else
			return(-1);	/* error */
	}

	*ptr = 0;
	return(n);
}
/*
 * Write "n" bytes to a descriptor.
 * Use in place of write() when fd is a stream socket.
 */

int
writen(fd, ptr, nbytes)
register int	fd;
register char	*ptr;
register int	nbytes;
{
	int	nleft, nwritten;

	nleft = nbytes;
	while (nleft > 0) {
		nwritten = write(fd, ptr, nleft);
		if (nwritten <= 0)
			return(nwritten);		/* error */

		nleft -= nwritten;
		ptr   += nwritten;
	}
	return(nbytes - nleft);
}

dowww(sd)
{
	/* process web request */
	int fd,n,i,ok;
	char *p, filename[256];
	

	while((n=readline(sd,buff,sizeof(buff))) > 0){
		buff[n]=0;
		if (n < 3) break;  /* end of html input */
		p = strstr(buff,"GET");
		if (p == NULL) continue;
		sscanf(p+4,"%s",filename);
		if (strcmp(filename,"/") == 0){
			/* feed em default page */
			strcpy(filename,Mypagefile);
		}
		/* try to open and give em what they want */
		writen(sd,MsgOK,strlen(MsgOK));
		printf("file %s\n",filename);
		ok=0;
		for(i=0; okfile[i]; i++ ){
		  /* restrict file access */
		  if (strcmp(okfile[i],filename)==0){ok=1; break;}
		}
		if (ok == 0){
			writen(sd,MsgNope,strlen(MsgNope));
			printf("access denied\n");
			continue;
		}
        	fd =open(filename+1,0);  /* open file for read */
		if (fd < 0) {
			close(fd);
			writen(sd,MsgNope,strlen(MsgNope));
			printf("file not found\n");
			continue;
		}
        	while( (n=read(fd,buff,sizeof(buff))) > 0){
			writen(sd,buff,n);
		}
		close(fd);
	}
	close(sd);
}
