/*
 *  web100srv [port]          v1   thd@ornl.gov
 *on thistle
 cc  -I/home/linux/web100/util/src  -o web100srv web100srv.c web100subs.c /home/linux/web100/util/src/web100lib.o
 *cc  -I/home/dunigan/linux/web100/util/src -o web100srv web100srv.c web100subs.c /home/dunigan/linux/web100/util/src/web100lib.o
 *   concurrent server
 * reap children, and thus watch out for EINTR
 *   tcp bandwidth tester from applet  Tcpbw100.java
 *  test 10s up and 10s down
 *  report web100 stats back to client 
 *  accept on server port, create 2nd and 3rd server port
 *     having server creates ports help if client is behind a firewall
 * might wish to add tcpwrappers to limit who can test....
 *  may need a better (hash) random (uncompressable) data stream
 */

#include        <stdio.h>
#include        <sys/types.h>
#include        <sys/socket.h>
#include 	<sys/time.h>
#include        <netinet/in.h>
#include        <arpa/inet.h>
#include        <netdb.h>
#include        <signal.h>
#include	"web100api.h"

#define LOGFILE "web100srv.log"
#define PORT 3005
#define RECLTH 8192
#define BUFFSIZE        8192
char buff[BUFFSIZE+1];
int window = 64000;
int randomize=0;
char *rmt_host;
double bwin, bwout;

#include <sys/errno.h>
extern int errno;
extern int *mibvalues;

main(argc, argv)
int	argc;
char	*argv[];
{
	int port = PORT;
	int reap(), ret=0;
	int n,sockfd, newsockfd, ctlsockfd, clilen, childpid, servlen, one=1;
	struct sockaddr_in	cli_addr, srv2_addr, srv3_addr, serv_addr;
	struct web100_agent *agent;
	struct web100_connection *conn;

	if (argc > 1) port = atoi(argv[1]);
	if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		err_sys("server: can't open stream socket");

	/*
	 * Bind our local address so that the client can send to us.
	 */

	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port        = htons(port);

	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));
	if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		err_sys("server: can't bind local address");

	setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &window, sizeof(window));
	setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &window, sizeof(window));
	listen(sockfd, 5);
	signal(SIGCLD,reap);		/* get rid of zombies */
	printf("server ready on port %d\n",port);
	  /* Initialize Web100 structures */
	agent=web100_Init();
		/*
		 * Wait for a connection from a client process.
		 */

	for(;;){
		int i;
		char *name, *ctime();
		struct hostent *hp;
		FILE *fp;
		clilen = sizeof(cli_addr);
		ctlsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
		if (ctlsockfd < 0){
			if (errno == EINTR) continue; /*sig child */
			err_sys("server: accept error");
		}
                i=time(0);
#ifdef NOPE
                hp = gethostbyaddr(&cli_addr.sin_addr,
                 sizeof(struct in_addr), AF_INET);
                name =  (hp == NULL) ? inet_ntoa(cli_addr.sin_addr) : hp->h_name;
#endif
		name = rmt_host = inet_ntoa(cli_addr.sin_addr);
                fp=fopen(LOGFILE,"a");
                fprintf(fp,"%15.15s  %s port %d\n",
                 ctime(&i)+4,name,ntohs(cli_addr.sin_port));
                fclose(fp);


		if (fork() == 0){ /* child */
			int sock2,sock3,port2,port3,bytes=0;
			int to,rec,minrtt,mywin,hiswin,rxmt,rtt,pktsout;
			double s,t,secs();

			close(sockfd);
			alarm(60);  /* die in 60 seconds */
			if ( (sock2 = socket(AF_INET, SOCK_STREAM, 0)) < 0)
				err_sys("server: can't open stream socket");
			bzero((char *) &srv2_addr, sizeof(srv2_addr));
			srv2_addr.sin_family = AF_INET;
			srv2_addr.sin_addr.s_addr = htonl(INADDR_ANY);
			srv2_addr.sin_port        = 0;  /* any */
		
			if (bind(sock2, (struct sockaddr *) &srv2_addr, sizeof(serv_addr)) < 0)
				err_sys("server: can't bind local address");

			setsockopt(sock2, SOL_SOCKET, SO_SNDBUF, &window, sizeof(window));
			setsockopt(sock2, SOL_SOCKET, SO_RCVBUF, &window, sizeof(window));
			listen(sock2, 2);
			/* send 2nd server port to client */
			clilen = sizeof(cli_addr);
			getsockname(sock2, (struct sockaddr *) &cli_addr, &clilen);
			port2 = ntohs(cli_addr.sin_port);
			if ( (sock3 = socket(AF_INET, SOCK_STREAM, 0)) < 0)
				err_sys("server: can't open stream socket");
			bzero((char *) &srv3_addr, sizeof(srv3_addr));
			srv3_addr.sin_family = AF_INET;
			srv3_addr.sin_addr.s_addr = htonl(INADDR_ANY);
			srv3_addr.sin_port        = 0;  /* any */
		
			if (bind(sock3, (struct sockaddr *) &srv3_addr, sizeof(serv_addr)) < 0)
				err_sys("server: can't bind local address");

			setsockopt(sock3, SOL_SOCKET, SO_SNDBUF, &window, sizeof(window));
			setsockopt(sock3, SOL_SOCKET, SO_RCVBUF, &window, sizeof(window));
			listen(sock3, 2);
			/* send server ports to client */
			clilen = sizeof(cli_addr);
			getsockname(sock3, (struct sockaddr *) &cli_addr, &clilen);
			port3 = ntohs(cli_addr.sin_port);
			printf("server ports %d %d\n",port2,port3);
			sprintf(buff,"%d %d",port2,port3);
			write(ctlsockfd,buff,strlen(buff));  
			/* ok, await for connect on 2nd port */
			newsockfd = accept(sock2, (struct sockaddr *) &cli_addr, &clilen);
			t=secs();
			while((n = read(newsockfd,buff,sizeof(buff))) > 0)
			   bytes += n;
			t=secs()-t;
			bwout = 8.e-3*bytes/t;
			sprintf(buff,"%6.0f Kbs outbound",bwout);
			printf("%s\n",buff);
			close(newsockfd);
			close(sock2);
			write(ctlsockfd,buff,strlen(buff)); /* send result */
			/* ok, await for connect on 3rd port and
    			 *  stream at him for 10 seconds
			 */
			clilen = sizeof(cli_addr);
			newsockfd = accept(sock3, (struct sockaddr *) &cli_addr, &clilen);
			bytes=0;
			t= secs();
			s= t +10.;
			while(secs() < s){ 
			  if (randomize) for(i=0;i<RECLTH;i++)
			    buff[i] = rand()>>24; /* defeat compression */
			  n=write(newsockfd,buff,RECLTH);
			  if (n < 0 ) break;
			  bytes += n;
			}
			shutdown(newsockfd,1);  /* end of write's */
			t=secs()-t;
			bwin = 8.e-3*bytes/t;
			conn = web100_connection_from_socket(agent, newsockfd);
			if((mibvalues = web100_Report(conn))==NULL) {
			  fprintf(stderr, "web100_Report returned NULL!\n");
			  exit(2);
			}
			ret = web100_Print1(mibvalues, ctlsockfd);
			/* now when client closes other end, read will fail */
			read(newsockfd,buff,1);  /* read fail/complete */
			/* send web100 results for newsockfd */
			close(newsockfd);
			close(sock3);
			close(ctlsockfd);
		/* get some web100 vars */
		web100_Logging(&to,&rec,&minrtt,&mywin,&hiswin,&rxmt,&rtt, &pktsout);
                fp=fopen(LOGFILE,"a");
                fprintf(fp,"%s iomhrRx3t %d %d Kbs %d %d B %d %d ms %d %d %d %d\n",
                 rmt_host,(int)bwin,(int)bwout,mywin,hiswin,minrtt,rtt,rxmt,rec,to,pktsout);
                fclose(fp);

			exit(0);
		}
		close(ctlsockfd);
	}

}

#include        <sys/wait.h>
void reap()
{
        /*
	 * avoid zombies, since we run forever
         * Use the wait3() system call with the WNOHANG option.
         */

        int             pid;
        int      status;

        while ( (pid = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0)
                ;
}

err_sys(s)
char *s;
{
	perror(s);
	exit(1);
}

double
secs()
{
        struct timeval ru;
        gettimeofday(&ru, (struct timezone *)0);
        return(ru.tv_sec + ((double)ru.tv_usec)/1000000);
}


